//
// dat2fl4.cpp
//
#define VERSION "8.1.2012"
//
// convert Dream hvxc dat dump file to fl4 format for decoding
//
#define CPR "(c) 2012 Dr.-Ing. Ingo D. Rullhusen <d01c@uni-bremen.de>"
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
//
//


#include <math.h>

#include <iostream>
#include <fstream>

using namespace std;


class bitstreamer
{
   public:
      bitstreamer() : nextBit(128), byte(0) {};

      void clear() { nextBit = 128; byte = 0; }

      void bitsOut( ofstream &ostream, unsigned long b, int nofbits )
      {
         unsigned long mask( 1 );
         mask <<= nofbits - 1;

         for ( int bn=0; bn<nofbits; ++bn )
         {
            bool bit = b & mask;

            byte |= ( bit ? nextBit : 0 );

            mask >>= 1;
            nextBit >>= 1;
            if ( nextBit == 0 )
            {
               ostream.write( (char*)&byte, sizeof(byte) );
               clear();
            }
         
         }

      }

      void padding( ofstream &ostream )
      {
         if ( nextBit < 128 )
         { ostream.write( (char*)&byte, sizeof(byte) ); } // null padding implicit
         clear();
      }

   private:
      short nextBit;
      unsigned char byte;
};


class bit_manipulator
{
   public:
      bit_manipulator( unsigned int n, unsigned char *data_ )
      : nbyte(n), nbit(8*n), data(data_) { clear(); }

      void setBit( unsigned int idx, bool bit )
      {
         int nby( idx/8 );
         int nbi( idx - nby*8 );

         if ( idx >= nextbit ) { nextbit = idx + 1; }

         unsigned char mask( 128 );
         mask >>= nbi;

         if ( bit ) { data[nby] |=  mask; }
         else       { data[nby] &= ~mask; }
      }

      bool getBit( unsigned int idx ) const
      {
         int nby( idx/8 );
         int nbi( idx - nby*8 );

         unsigned char mask( 128 );
         mask >>= nbi;

         return( data[nby] & mask );
      }

      void appendBits( unsigned int ui, int nofbits )
      {
         unsigned int mask( 1 );
         mask <<= nofbits - 1;

         unsigned int idx1( nextbit ), idx2( nextbit+nofbits );
         for ( unsigned int i=idx1; i<idx2; ++i )
         {
             setBit( i, ui & mask );
             mask >>= 1;
         }
      }

      unsigned int nextBitIdx() { return(nextbit); }

      void setNextBitIdx( unsigned int idx ) { nextbit = idx; }

      void clear()
      {
         for ( unsigned int i=0; i<nbyte; ++i ) { data[i] = 0; }
         nextbit = 0;
      }

   private:
      unsigned int nbyte, nbit, nextbit;
      unsigned char *data;
};


//
// ETSI ES 201 980 V3.1.1 (2009-08): clause 5.5.5, fig. 13
void deinterlace( bit_manipulator &bMout,
                  const bit_manipulator &bMin, int N1, int N2 )
{
   int n, n1, n2, i0, i1, N;
   n = n1 = n2 = i0 = i1 = 0;

   N = N1 + N2;

   bMout.clear();
   do
   {
      if ( i0 <= i1 )
      {
         bMout.setBit( n1, bMin.getBit(n) );
         ++n1;

         i0 += N;
      }
      else
      {
         bMout.setBit( N1 + n2, bMin.getBit(n) );
         ++n2;
      }

      i1 += N1;
      ++n;

   } while ( n < N );
}


//
// ETSI TS 101 980 V1.1.1 (2001-09): clause 5.5.4.2
// ISO/IEC 14496-3 Subpart 1:1998: clause 2
void ESC2k_to_plain( bit_manipulator &bMout,
                     const bit_manipulator &bMin, bool CRC )
{
   int n( 0 );
   bMout.clear();

// VUV
   bool vuv( false );
   vuv = vuv || bMin.getBit(n);
   bMout.setBit( 18, bMin.getBit(n) ); ++n;
   vuv = vuv || bMin.getBit(n);
   bMout.setBit( 19, bMin.getBit(n) ); ++n;

// LSP4 0
   bMout.setBit( 17, bMin.getBit(n) ); ++n;

   if ( vuv )
   {
// SE_gain 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 35+i, bMin.getBit(n) ); ++n; }

// LSP1 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 0+i, bMin.getBit(n) ); ++n; }

// Pitch 6...1
      for ( int i=0; i<6; ++i )
      { bMout.setBit( 20+i, bMin.getBit(n) ); ++n; }

// LSP2 6
      bMout.setBit( 5, bMin.getBit(n) ); ++n;

// LSP3 4
      bMout.setBit( 12, bMin.getBit(n) ); ++n;

// LSP2 5
      bMout.setBit( 6, bMin.getBit(n) ); ++n;

// ESC0 CRC
      if ( CRC ) n += 6;

//
// SE_shape1 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 27+i, bMin.getBit(n) ); ++n; }

// ESC1 CRC
      if ( CRC ) n += 1;

//
// SE_shape2 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 31+i, bMin.getBit(n) ); ++n; }

// ESC2 CRC
      if ( CRC ) n += 1;

//
// LSP2 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 7+i, bMin.getBit(n) ); ++n; }

// LSP3 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 13+i, bMin.getBit(n) ); ++n; }

// Pitch 0
      bMout.setBit( 26, bMin.getBit(n) ); ++n;
   }
   else
   {
// VX_gain1[0] 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 26+i, bMin.getBit(n) ); ++n; }

// VX_gain1[1] 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 36+i, bMin.getBit(n) ); ++n; }

// LSP1 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 0+i, bMin.getBit(n) ); ++n; }

// LSP2 6...3
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 5+i, bMin.getBit(n) ); ++n; }

// LSP3 4...3
      for ( int i=0; i<2; ++i )
      { bMout.setBit( 12+i, bMin.getBit(n) ); ++n; }

// ESC0 CRC
      if ( CRC ) n += 6;

//
// LSP2 2...0
      for ( int i=0; i<3; ++i )
      { bMout.setBit( 9+i, bMin.getBit(n) ); ++n; }

// LSP3 2
      bMout.setBit( 14, bMin.getBit(n) ); ++n;

// ESC1 CRC
      if ( CRC ) n += 1;

//
// LSP3 1...0
      for ( int i=0; i<2; ++i )
      { bMout.setBit( 15+i, bMin.getBit(n) ); ++n; }

// VX_shape1[0] 5...4
      for ( int i=0; i<2; ++i )
      { bMout.setBit( 20+i, bMin.getBit(n) ); ++n; }

// ESC2 CRC
      if ( CRC ) n += 1;

//
// VX_shape1[0] 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 22+i, bMin.getBit(n) ); ++n; }

// VX_shape1[1] 5...0
      for ( int i=0; i<6; ++i )
      { bMout.setBit( 30+i, bMin.getBit(n) ); ++n; }
   }

}

//
// ETSI TS 101 980 V1.1.1 (2001-09): clause 5.5.4.2
// ISO/IEC 14496-3 Subpart 1:1998: clause 2
void ESC4k_to_plain( bit_manipulator &bMout,
                     const bit_manipulator &bMin, bool CRC )
{
   int n( 0 );
   bMout.clear();

// VUV
   bool vuv( false );
   vuv = vuv || bMin.getBit(n);
   bMout.setBit( 18, bMin.getBit(n) ); ++n;
   vuv = vuv || bMin.getBit(n);
   bMout.setBit( 19, bMin.getBit(n) ); ++n;

// LSP4 0
   bMout.setBit( 17, bMin.getBit(n) ); ++n;

   if ( vuv )
   {
// SE_gain 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 35+i, bMin.getBit(n) ); ++n; }

// LSP1 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 0+i, bMin.getBit(n) ); ++n; }

// Pitch 6...1
      for ( int i=0; i<6; ++i )
      { bMout.setBit( 20+i, bMin.getBit(n) ); ++n; }

// LSP2 6...3
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 5+i, bMin.getBit(n) ); ++n; }

// SE_shape3 6...2
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 48+i, bMin.getBit(n) ); ++n; }

// LSP3 4
      bMout.setBit( 12, bMin.getBit(n) ); ++n;

// LSP5 7
      bMout.setBit( 40, bMin.getBit(n) ); ++n;

// SE_shape4 9
      bMout.setBit( 55, bMin.getBit(n) ); ++n;

// SE_shape5 8
      bMout.setBit( 65, bMin.getBit(n) ); ++n;

// SE_shape6 5
      bMout.setBit( 74, bMin.getBit(n) ); ++n;

// ESC0 CRC 5...0
      if ( CRC ) n += 6;

//
// SE_shape4 8...0
      for ( int i=0; i<9; ++i )
      { bMout.setBit( 56+i, bMin.getBit(n) ); ++n; }

// SE_shape5 7...0
      for ( int i=0; i<8; ++i )
      { bMout.setBit( 66+i, bMin.getBit(n) ); ++n; }

// SE_shape6 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 75+i, bMin.getBit(n) ); ++n; }

// ESC1 CRC 4...0
      if ( CRC ) n += 5;

//
// SE_shape1 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 27+i, bMin.getBit(n) ); ++n; }

// ESC2 CRC 0
      if ( CRC ) n += 1;

//
// SE_shape2 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 31+i, bMin.getBit(n) ); ++n; }

// ESC3 CRC 0
      if ( CRC ) n += 1;

//
// LSP2 2...0
      for ( int i=0; i<3; ++i )
      { bMout.setBit( 9+i, bMin.getBit(n) ); ++n; }

// LSP3 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 13+i, bMin.getBit(n) ); ++n; }

// LSP5 6...0
      for ( int i=0; i<7; ++i )
      { bMout.setBit( 41+i, bMin.getBit(n) ); ++n; }

// Pitch 0
      bMout.setBit( 26, bMin.getBit(n) ); ++n;

// SE_shape3 1...0
      for ( int i=0; i<2; ++i )
      { bMout.setBit( 53+i, bMin.getBit(n) ); ++n; }
   }
   else
   {
// VX_gain1[0] 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 26+i, bMin.getBit(n) ); ++n; }

// VX_gain1[1] 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 36+i, bMin.getBit(n) ); ++n; }

// LSP1 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 0+i, bMin.getBit(n) ); ++n; }

// LSP2 6...3
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 5+i, bMin.getBit(n) ); ++n; }

// LSP3 4
      bMout.setBit( 12, bMin.getBit(n) ); ++n;

// LSP5 7
      bMout.setBit( 40, bMin.getBit(n) ); ++n;

// VX_gain2[0] 2...0
      for ( int i=0; i<3; ++i )
      { bMout.setBit( 53+i, bMin.getBit(n) ); ++n; }

// VX_gain2[1] 2...0
      for ( int i=0; i<3; ++i )
      { bMout.setBit( 61+i, bMin.getBit(n) ); ++n; }

// VX_gain2[2] 2...0
      for ( int i=0; i<3; ++i )
      { bMout.setBit( 69+i, bMin.getBit(n) ); ++n; }

// VX_gain2[3] 2...1
      for ( int i=0; i<2; ++i )
      { bMout.setBit( 77+i, bMin.getBit(n) ); ++n; }

// ESC0 CRC 5...0
      if ( CRC ) n += 6;

//
// VX_gain2[3] 0
      bMout.setBit( 79, bMin.getBit(n) ); ++n;

// LSP2 2...0
      for ( int i=0; i<3; ++i )
      { bMout.setBit( 9+i, bMin.getBit(n) ); ++n; }

// LSP3 3...0
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 13+i, bMin.getBit(n) ); ++n; }

// LSP5 6...0
      for ( int i=0; i<7; ++i )
      { bMout.setBit( 41+i, bMin.getBit(n) ); ++n; }

// VX_shape1[0] 5...0
      for ( int i=0; i<6; ++i )
      { bMout.setBit( 20+i, bMin.getBit(n) ); ++n; }

// VX_shape1[1] 5
      bMout.setBit( 30, bMin.getBit(n) ); ++n;

// ESC1 CRC 4...0
      if ( CRC ) n += 5;

//
// VX_shape1[1] 4...1
      for ( int i=0; i<4; ++i )
      { bMout.setBit( 31+i, bMin.getBit(n) ); ++n; }

// ESC2 CRC 0
      if ( CRC ) n += 1;

//
// VX_shape1[1] 0
      bMout.setBit( 35, bMin.getBit(n) ); ++n;

// VX_shape2[0] 4...2
      for ( int i=0; i<3; ++i )
      { bMout.setBit( 48+i, bMin.getBit(n) ); ++n; }

// ESC3 CRC 0
      if ( CRC ) n += 1;

//
// VX_shape2[0] 1...0
      for ( int i=0; i<2; ++i )
      { bMout.setBit( 51+i, bMin.getBit(n) ); ++n; }

// VX_shape2[1] 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 56+i, bMin.getBit(n) ); ++n; }

// VX_shape2[2] 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 64+i, bMin.getBit(n) ); ++n; }

// VX_shape2[3] 4...0
      for ( int i=0; i<5; ++i )
      { bMout.setBit( 72+i, bMin.getBit(n) ); ++n; }
   }

}


int main( int argc, char *argv[] )
{
   unsigned long ul, ltmp;
   ifstream ifile;
   ofstream ofile;
   bitstreamer bS;

   const int bitbuffersize = 64; // size for file header
   unsigned char bitbuffer_1[bitbuffersize];
   bit_manipulator bM_1( bitbuffersize, bitbuffer_1 );
   unsigned char bitbuffer_2[bitbuffersize];
   bit_manipulator bM_2( bitbuffersize, bitbuffer_2 );

   if ( argc != 3 )
   {
      cout << "dat2fl4 (version " << VERSION << ")" << endl;
      cout << " Copyright " << CPR << endl;
      cout << " dat2fl4 comes with ABSOLUTELY NO WARRANTY;" << endl;
      cout << " for details see GNU GENERAL PUBLIC LICENSE. This is free software," << endl;
      cout << " and you are welcome to redistribute it under certain conditions;" << endl;
      cout << " see GNU GENERAL PUBLIC LICENSE for details." << endl << endl;
      cout << "usage:" << endl << endl;
      cout << " dat2fl4 <.dat input file> <.fl4 output file>" << endl << endl;
      cout << " <.dat input  file> Dream dump .dat file with DRM HVXC frames" << endl;
      cout << " <.fl4 output file> .fl4 file with plain HVXC frames" << endl;
      cout << endl;

      return(1);
   }


   ifile.open(argv[1], ios::in|ios::binary );
   if ( ! ifile.is_open() )
   {
      cerr << "Error: cannot open \"" << argv[1] << "\"!" << endl;
      return(1);
   }
   ofile.open(argv[2], ios::out|ios::binary );
   if ( ! ofile.is_open() )
   {
      cerr << "Error: cannot open \"" << argv[2] << "\"!" << endl;
      return(1);
   }

// setup
   unsigned int fsize_init;
   ifile.read( (char*)&fsize_init, sizeof(fsize_init) );
   ifile.seekg( 0, ios::beg );

   unsigned long samplerate, bitrate, bpfr, ofsize;
   int N1, N2;
   bool CRC;

   samplerate = 8000;
   switch (fsize_init)
   {
      case  5:
      {
         bitrate = 2000; bpfr = 40; CRC = false;
         ofsize = 5;  N1 = 22;
         break;
      }
      case  6:
      {
         bitrate = 2000; bpfr = 48; CRC = true;
         ofsize = 5;  N1 = 28;
         break;
      }
      case 10:
      {
         bitrate = 4000; bpfr = 80; CRC = false;
         ofsize = 10; N1 = 33;
         break;
      }
      case 12:
      {
         bitrate = 4000; bpfr = 93; CRC = true;
         ofsize = 10; N1 = 39;
         break;
      }
      default:
      {
         cerr << "Error: illegal frame size!\n" << endl;
         return(1);
      }
   }
   N2 = bpfr - N1;

   if (CRC) { cerr << "Warning: CRC ignored." << endl; }


   ofile << ".mp4\n";
   bS.bitsOut( ofile, 0, 8 );

// .fl4 stream header in bitbuffer
   bM_1.clear();
// advanceODescr
//   ul = 21;
//   bM_1.appendBits( ul, 32 ); // ODLength [later]
   ul = 1;
   bM_1.appendBits( ul, 10 ); // ODDescrId
   ul = 1;
   bM_1.appendBits( ul, 5 ); // streamCount
   ul = 0;
   bM_1.appendBits( ul, 1 ); // extensionFlag

// advanceESDescr
   ul = 0;
   bM_1.appendBits( ul, 5 ); // ESNumber
   ltmp = ((ul==0)?0:1);
   bM_1.appendBits( ltmp, 1 ); // streamDependence
   if ( ltmp != 0 )
   { ltmp = ( ul>0 ? ul : 0 );
     bM_1.appendBits( ul, 5 ); // dependsOn_Es_number
   }
   ltmp = 0;
   bM_1.appendBits( ltmp, 1 ); // URLFlag
   bM_1.appendBits( ltmp, 1 ); // extensFlag

// -> DecConfigDescr
   ul = 0;
   bM_1.appendBits( ul, 8 ); // profileAndLevelIndication
   ul = 5;   // audio stream
   bM_1.appendBits( ul, 6 ); // streamType
   ul = 0;
   bM_1.appendBits( ul, 1 ); // upsteam
   ul = 1;
   bM_1.appendBits( ul, 1 ); // specificInfoFlag
   ul = ofsize;
   bM_1.appendBits( ul, 24 ); // bufferSizeDB
   ul = bitrate;
   bM_1.appendBits( ul, 32 ); // maxBitrate
   bM_1.appendBits( ul, 32 ); // avgBitrate
   ul = 3;
   bM_1.appendBits( ul, 8 ); // specificInfoLength [audioDecoderType...extensionFlag]
   ul = 9;   // HVXC Object
   bM_1.appendBits( ul, 5 ); // audioDecoderType
   ul = 0xb;   // 8kHz
   bM_1.appendBits( ul, 4 ); // samplingFreqencyIndex
   if ( ul == 0xf )
   { ul = samplerate;
     bM_1.appendBits( ul, 24 ); // samplingFrequency
   }
   ul = 1;
   bM_1.appendBits( ul, 4 ); // channelConfiguration

   ul = 1;
   bM_1.appendBits( ul, 1 ); // isBaseLayer
   ul = 0;
   bM_1.appendBits( ul, 1 ); // HVXCvarMode
   ul = ( bitrate == 2000 ? 0 : 1 );
   bM_1.appendBits( ul, 2 ); // HVXCrateMode
   ul = 0;
   bM_1.appendBits( ul, 1 ); // extensionFlag

///   ul = 0;
///   bM_1.appendBits( ul, 1 ); // epConfig

// -> ALConfigDescriptor
   ul = 1;
   bM_1.appendBits( ul, 1 ); // useAccessUnitStartFlag
   ul = 1;
   bM_1.appendBits( ul, 1 ); // useAccessUnitEndFlag
   ul = 0;
   bM_1.appendBits( ul, 1 ); // useRandomAccessPointFlag
   bM_1.appendBits( ul, 1 ); // usePaddingFlag
   bM_1.appendBits( ul, 4 ); // seqNumLength


// .fl4 stream header output
   ltmp = (int)ceil( bM_1.nextBitIdx()/8.0 );
   bS.bitsOut( ofile, ltmp, 32 );

   for ( unsigned int i=0; i<ltmp; ++i )
   { bS.bitsOut( ofile, bitbuffer_1[i], 8 ); }  // padding included


// frame by frame
   for ( int fno=0; true; ++fno )
   {
      unsigned int fsize;
      ifile.read( (char*)&fsize, sizeof(fsize) );
      if ( ifile.fail() ) { break; }

      if ( fsize != fsize_init )
      {
         cerr << "Error: coding parameter changed in frame "
              << fno
              << ", stop!" << endl;
         return(1);
      }

      bM_1.clear();
      ifile.read( (char*)bitbuffer_1, fsize );
      if ( ifile.fail() ) { break; }
      bM_1.setNextBitIdx( bpfr );


// .fl4 frame header
      ul = 0; // stream index
      bS.bitsOut( ofile, ul, 8 );
      ul = ofsize; // max bytes;
      bS.bitsOut( ofile, ul, 8 );
      ul = 1; // AUStartFlag [first subframe max. 255 byte]
      bS.bitsOut( ofile, ul, 1 );
      ul = 1; // AUEndFlag [last subframe < 255 byte]
      bS.bitsOut( ofile, ul, 1 );
      bS.padding( ofile ); // padding bits


      deinterlace( bM_2, bM_1, N1, N2 );

// place CRC operation here

      if ( bitrate == 2000 )
      { ESC2k_to_plain( bM_1, bM_2, CRC ); }
      else
      { ESC4k_to_plain( bM_1, bM_2, CRC ); }


      for ( unsigned int i=0; i<ofsize; ++i )
      { bS.bitsOut( ofile, bitbuffer_1[i], 8 ); }
   
   }


   bS.padding( ofile );

   ofile.close();
   ifile.close();

   return(0);
}
